import json

import dash_colorscales as dcs
import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
import plotly
from pkg_resources import parse_version

import dtale.dash_application.custom_geojson as custom_geojson
from dtale.charts.utils import YAXIS_CHARTS, ZAXIS_CHARTS, find_group_vals
from dtale.dash_application.layout.utils import build_input, build_option
from dtale.utils import (
    ChartBuildingError,
    classify_type,
    dict_merge,
    flatten_lists,
    get_dtypes,
    inner_build_query,
    is_app_root_defined,
    make_list,
)


def test_plotly_version(version_num):
    if "unknown" in plotly.__version__:
        return True
    return parse_version(plotly.__version__) >= parse_version(version_num)


def base_layout(github_fork, app_root, **kwargs):
    """
    Base layout to be returned by :meth:`dtale.dash_application.views.DtaleDash.interpolate_index`

    :param github_fork: `True` if "Fork me on Github" banner should be displayed, `False` otherwise
    :type github_fork: bool
    :param kwargs: Optional keyword arguments to be passed to 'dash.Dash.interplolate_index'
    :type kwargs: dict
    :return: HTML
    :rtype: str
    """
    back_to_data_padding, github_fork_html, webroot_html = ("", "", "")
    if github_fork:
        back_to_data_padding = "padding-right: 125px"
        github_fork_html = """
            <span id="forkongithub">
                <a href="https://github.com/man-group/dtale">Fork me on GitHub</a>
            </span>
        """
    favicon_path = "../../images/favicon.png"
    if is_app_root_defined(app_root):
        webroot_html = """
        <script type="text/javascript">
            window.resourceBaseUrl = '{app_root}';
        </script>
        """.format(
            app_root=app_root
        )
        favicon_path = "{}/images/favicon.png".format(app_root)
    return """
        <!DOCTYPE html>
        <html>
            <head>
                {webroot_html}
                {metas}
                <title>D-Tale Charts</title>
                <link rel="shortcut icon" href="{favicon_path}">
                {css}
            </head>
            <body>
                {github_fork_html}
                <div class="container-fluid charts">
                    <div class="row" style="margin: 0">
                        <div class="col-auto">
                            <header>
                                <span class="title-font">D-TALE</span>
                                <span style="font-size: 16px" class="pl-4">Charts</span>
                            </header>
                        </div>
                        <div class="col"></div>
                        <div class="col-auto mt-4" style="{back_to_data_padding}">
                            <a href="#" onclick="javascript:backToData()">
                                <i class="fas fa-th mr-4"></i>
                                <span>Back To Data</span>
                            </a>
                        </div>
                    </div>
                    {app_entry}
                </div>
                <footer>
                    {config}
                    {scripts}
                    <script type="text/javascript">
                        const pathSegs = window.location.pathname.split('/');
                        const dataId = pathSegs[pathSegs.length - 1];
                        const backToData = () => window.open('{app_root}/dtale/main/' + dataId);
                        const goToLegacy = () => location.replace('{app_root}/dtale/popup/charts/' + dataId);
                    </script>
                    {renderer}
                    {css}
                </footer>
            </body>
        </html>
    """.format(
        metas=kwargs["metas"],
        css=kwargs["css"],
        app_entry=kwargs["app_entry"],
        config=kwargs["config"],
        scripts=kwargs["scripts"],
        renderer=kwargs["renderer"],
        back_to_data_padding=back_to_data_padding,
        webroot_html=webroot_html,
        github_fork_html=github_fork_html,
        app_root=app_root if is_app_root_defined(app_root) else "",
        favicon_path=favicon_path,
    )


def build_tab(label, value, additional_style=None, **kwargs):
    """
    Builds a :dash:`dash_core_components.Tab <dash-core-components/tab>` with standard styling settings.
    """
    base_style = {"borderBottom": "1px solid #d6d6d6", "padding": "6px"}
    return dcc.Tab(
        label=label,
        value=value,
        style=dict_merge(base_style, {"fontWeight": "bold"}, additional_style or {}),
        disabled_style=dict_merge(
            base_style,
            {
                "fontWeight": "bold",
                "backgroundColor": "LightGray",
                "color": "black",
                "cursor": "not-allowed",
            },
            additional_style or {},
        ),
        selected_style=dict_merge(
            base_style,
            {
                "borderTop": "1px solid #d6d6d6",
                "backgroundColor": "#2a91d1",
                "color": "white",
            },
            additional_style or {},
        ),
        **kwargs
    )


CHARTS = [
    dict(value="line"),
    dict(value="bar"),
    dict(value="scatter"),
    dict(value="pie"),
    dict(value="wordcloud"),
    dict(value="heatmap"),
    dict(value="3d_scatter", label="3D Scatter"),
    dict(value="surface"),
    dict(value="maps", label="Maps"),
]
CHART_INPUT_SETTINGS = {
    "line": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "bar": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "scatter": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "pie": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "wordcloud": dict(
        x=dict(type="single"),
        y=dict(type="multi"),
        z=dict(display=False),
        group=dict(display=True, type="single"),
    ),
    "heatmap": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=False),
    ),
    "3d_scatter": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=True),
    ),
    "surface": dict(
        x=dict(type="single"),
        y=dict(type="single"),
        z=dict(display=True, type="single"),
        group=dict(display=False),
    ),
    "maps": dict(
        x=dict(display=False),
        y=dict(display=False),
        z=dict(display=False),
        group=dict(display=False),
        map_group=dict(display=True),
    ),
}
AGGS = dict(
    raw="No Aggregation",
    count="Count",
    nunique="Unique Count",
    sum="Sum",
    mean="Mean",
    rolling="Rolling",
    corr="Correlation",
    first="First",
    last="Last",
    median="Median",
    min="Minimum",
    max="Maximum",
    std="Standard Deviation",
    var="Variance",
    mad="Mean Absolute Deviation",
    prod="Product of All Items",
    pctct="Percentage Count",
    pctsum="Percentage Sum",
)
FREQS = ["H", "H2", "WD", "D", "W", "M", "Q", "Y"]
FREQ_LABELS = dict(
    H="Hourly",
    H2="Hour",
    WD="Weekday",
    W="Weekly",
    M="Monthly",
    Q="Quarterly",
    Y="Yearly",
)

MAP_TYPES = [
    dict(value="choropleth", image=True),
    dict(value="scattergeo", label="ScatterGeo", image=True),
    dict(value="mapbox"),
]
SCOPES = ["world", "usa", "europe", "asia", "africa", "north america", "south america"]
PROJECTIONS = [
    "equirectangular",
    "mercator",
    "orthographic",
    "natural earth",
    "kavrayskiy7",
    "miller",
    "robinson",
    "eckert4",
    "azimuthal equal area",
    "azimuthal equidistant",
    "conic equal area",
    "conic conformal",
    "conic equidistant",
    "gnomonic",
    "stereographic",
    "mollweide",
    "hammer",
    "transverse mercator",
    "albers usa",
    "winkel tripel",
    "aitoff",
    "sinusoidal",
]


def build_img_src(proj, img_type="projections"):
    return "../images/{}/{}.png".format(img_type, "_".join(proj.split(" ")))


def build_proj_hover_children(proj):
    if proj is None:
        return None
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            [html.Div(proj), html.Img(src=build_img_src(proj))],
            className="hoverable__content",
            style=dict(width="auto"),
        ),
    ]


def build_proj_hover(proj):
    return html.Span(
        [
            "Projection",
            html.Div(
                build_proj_hover_children(proj),
                className="ml-3 hoverable",
                style=dict(display="none")
                if proj is None
                else dict(borderBottom="none"),
                id="proj-hover",
            ),
        ],
        className="input-group-addon",
    )


def build_mapbox_token_children():
    from dtale.charts.utils import get_mapbox_token

    msg = "To access additional styles enter a token here..."
    if get_mapbox_token() is None:
        msg = "Change your token here..."
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            [
                html.Span("Mapbox Access Token:"),
                dcc.Input(
                    id="mapbox-token-input",
                    type="text",
                    placeholder=msg,
                    className="form-control",
                    value="",
                    style={"lineHeight": "inherit"},
                ),
            ],
            className="hoverable__content",
            style=dict(width="20em", right="-1.45em"),
        ),
    ]


def build_mapbox_token_hover():
    return html.Span(
        [
            "Style",
            html.Div(
                build_mapbox_token_children(),
                className="ml-3 hoverable",
                style=dict(borderBottom="none"),
                id="token-hover",
            ),
        ],
        className="input-group-addon",
    )


LOC_MODE_INFO = {
    "ISO-3": dict(
        url=html.A(
            "ISO-3",
            href="https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3",
            target="_blank",
        ),
        examples=["USA (United States)", "CAN (Canada)", "GBR (United Kingdom)"],
    ),
    "USA-states": dict(
        url=html.A(
            "USA-states (use ISO)",
            href="https://en.wikipedia.org/wiki/List_of_U.S._state_abbreviations",
            target="_blank",
        ),
        examples=["NY (New York)", "CA (California)", "MA (Massachusetts)"],
    ),
    "country names": dict(
        url=html.A(
            "country names (case-insensitive)",
            href="https://simple.wikipedia.org/wiki/List_of_countries",
            target="_blank",
        ),
        examples=["United States", "United Kingdom", "Germany"],
    ),
    "geojson-id": dict(label="Custom GeoJSON", desc="Load custom geojson into D-Tale"),
}


def build_loc_mode_hover_children(loc_mode):
    if loc_mode is None:
        return None
    loc_mode_cfg = LOC_MODE_INFO[loc_mode]
    url, examples, desc = (loc_mode_cfg.get(p) for p in ["url", "examples", "desc"])
    body = []
    if url is not None:
        body.append(
            html.Div([html.Span("View", className="mr-3"), loc_mode_cfg["url"]])
        )
    if examples is not None:
        body.append(
            html.Ul(
                [
                    html.Li(e, style={"listStyleType": "disc"}, className="mb-3")
                    for e in loc_mode_cfg["examples"]
                ],
                className="pt-3 mb-0",
            )
        )
    if desc is not None:
        body.append(html.Span(desc))
    return [
        html.I(className="ico-help-outline", style=dict(color="white")),
        html.Div(
            body,
            className="hoverable__content build-code",
            style=dict(width="auto", whiteSpace="nowrap", left="-2em", top="auto"),
        ),
    ]


def build_loc_mode_hover(loc_mode):
    return html.Span(
        [
            html.Span("Location Mode", style=dict(whiteSpace="pre-line")),
            html.Div(
                build_loc_mode_hover_children(loc_mode),
                className="ml-3 hoverable",
                style=dict(display="none")
                if loc_mode is None
                else dict(borderBottom="none"),
                id="loc-mode-hover",
            ),
        ],
        className="input-group-addon pt-1 pb-0",
    )


JET = ["#000083", "#003CAA", "#05FFFF", "#FFFF00", "#FA0000", "#800000"]
REDS = ["#fff5f0", "#fdcab4", "#fc8a6a", "#f24632", "#bc141a", "#67000d"]
ANIMATION_CHARTS = ["line"]
ANIMATE_BY_CHARTS = ["bar", "3d_scatter", "maps"]


def show_input_handler(chart_type):
    settings = CHART_INPUT_SETTINGS.get(chart_type or "line") or {}

    def _show_input(input_id, input_type="single"):
        cfg = settings.get(input_id, {})
        return cfg.get("display", True) and cfg.get("type", "single") == input_type

    return _show_input


def show_group_input(inputs, group_cols=None):
    chart_type = inputs.get("chart_type")
    if show_input_handler(chart_type)("group"):
        return len(group_cols or make_list(inputs.get("group")))
    elif show_input_handler(chart_type)("map_group"):
        return len(group_cols or make_list(inputs.get("map_group")))
    return False


def update_label_for_freq(val):
    """
    Formats sub-values contained within 'val' to display date frequencies if included.
        - (val=['a', 'b', 'c']) => 'a, b, c'
        - (val=['a|H', 'b', 'c']) => 'a (Hour), b, c'
    """

    def _freq_handler(sub_val):
        for freq in FREQS:
            if sub_val.endswith("|{}".format(freq)):
                col, freq = sub_val.split("|")
                if freq in FREQ_LABELS:
                    return "{} ({})".format(col, FREQ_LABELS[freq])
                return col
        return sub_val

    return ", ".join([_freq_handler(sub_val) for sub_val in make_list(val)])


def build_error(error, tb):
    """
    Returns error/traceback information in standard component with styling

    :param error: execption message
    :type error: str
    :param tb: tracebackF
    :type tb: str
    :return: error component
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    if isinstance(error, ChartBuildingError):
        if error.details:
            tb = error.details
        error = error.error
    return html.Div(
        [
            html.I(className="ico-error"),
            html.Span(str(error)),
            html.Div(html.Pre(str(tb)), className="traceback"),
        ],
        className="dtale-alert alert alert-danger",
    )


def build_cols(cols, dtypes):
    """
    Helper function to add additional column entries for columns of type datetime so that users can make use of
    different frequencies of dates.  For example, hour, weekday, month, quarter, year

    :param cols: columns in dataframe
    :type cols: list of strings
    :param dtypes: datatypes of columns in dataframe
    :type dtypes: dict
    :return: generator or columns + any additional (datetime column + frequency) options
    """
    for c in cols:
        if classify_type(dtypes[c]) == "D":
            for freq in FREQS:
                if freq in FREQ_LABELS:
                    yield "{}|{}".format(c, freq), "{} ({})".format(
                        c, FREQ_LABELS[freq]
                    )
                else:
                    yield c, c
        else:
            yield c, c


def build_selections(*args):
    """
    simple helper function to build a single level list of values based on variable number of inputs which could be
    equal to None.
    """
    return flatten_lists([[] if a is None else make_list(a) for a in args])


def build_input_options(df, **inputs):
    """
    Builds dropdown options for (X, Y, Z, Group, Barsort & Y-Axis Ranges) with filtering based on currently selected
    values for the following inputs: x, y, z, group.
    """
    chart_type, x, y, z, group = (
        inputs.get(p) for p in ["chart_type", "x", "y", "z", "group"]
    )
    col_opts = list(build_cols(df.columns, get_dtypes(df)))
    group_val, z_val = (None, z) if chart_type in ZAXIS_CHARTS else (group, None)
    x_options = [
        build_option(c, l)
        for c, l in col_opts
        if c not in build_selections(y, z_val, group_val)
    ]
    y_filter = build_selections(x, group_val, z_val)
    y_multi_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    y_single_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    z_options = [
        build_option(c)
        for c in df.columns
        if c not in build_selections(x, y, group_val)
    ]
    group_options = [
        build_option(c, l)
        for c, l in col_opts
        if c not in build_selections(x, y, z_val)
    ]
    barsort_options = [build_option(o) for o in build_selections(x, y)]
    yaxis_options = [build_option(y2) for y2 in y or []]

    return (
        x_options,
        y_multi_options,
        y_single_options,
        z_options,
        group_options,
        barsort_options,
        yaxis_options,
    )


def build_map_options(
    df, type="choropleth", loc=None, lat=None, lon=None, map_val=None
):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    float_cols, str_cols, num_cols = [], [], []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification == "S":
            str_cols.append(c)
            continue
        if classification in ["F", "I"]:
            num_cols.append(c)
            if classification == "F":
                float_cols.append(c)

    lat_options = [
        build_option(c) for c in float_cols if c not in build_selections(lon, map_val)
    ]
    lon_options = [
        build_option(c) for c in float_cols if c not in build_selections(lat, map_val)
    ]
    loc_options = [
        build_option(c) for c in str_cols if c not in build_selections(map_val)
    ]

    if type == "choropleth":
        val_options = [
            build_option(c) for c in num_cols if c not in build_selections(loc)
        ]
    else:
        val_options = [
            build_option(c) for c in num_cols if c not in build_selections(lon, lat)
        ]
    return loc_options, lat_options, lon_options, val_options


def build_mapbox_style_options():
    from dtale.charts.utils import get_mapbox_token

    free_styles = [
        "open-street-map",
        "carto-positron",
        "carto-darkmatter",
        "stamen-terrain",
        "stamen-toner",
        "stamen-watercolor",
    ]
    token_styles = [
        "basic",
        "streets",
        "outdoors",
        "light",
        "dark",
        "satellite",
        "satellite-streets",
    ]
    styles = free_styles
    if get_mapbox_token() is not None:
        styles += token_styles
    return [build_option(v) for v in styles]


def bar_input_style(**inputs):
    """
    Sets display CSS property for bar chart inputs
    """
    chart_type, group_col = (inputs.get(p) for p in ["chart_type", "group"])
    show_bar = chart_type == "bar"
    show_barsort = show_bar and group_col is None
    return (
        dict(display="block" if show_bar else "none"),
        dict(display="block" if show_barsort else "none"),
    )


def colorscale_input_style(**inputs):
    return dict(
        display="block" if inputs.get("chart_type") in ["heatmap", "maps"] else "none"
    )


def animate_styles(df, **inputs):
    chart_type, agg, cpg = (inputs.get(p) for p in ["chart_type", "agg", "cpg"])
    opts = []
    if cpg or agg in ["pctsum", "pctct"]:
        return dict(display="none"), dict(display="none"), opts
    if chart_type in ANIMATION_CHARTS:
        return dict(display="block"), dict(display="none"), opts
    if chart_type in ANIMATE_BY_CHARTS:
        opts = [build_option(v, l) for v, l in build_cols(df.columns, get_dtypes(df))]
    if len(opts):
        return dict(display="none"), dict(display="block"), opts
    return dict(display="none"), dict(display="none"), []


def show_chart_per_group(**inputs):
    """
    Boolean function to determine whether "Chart Per Group" toggle should be displayed or not
    """
    [chart_type, group] = [inputs.get(p) for p in ["chart_type", "group"]]
    invalid_type = chart_type in ["pie", "wordcloud", "maps"]
    return (
        show_input_handler(chart_type)("group")
        and len(group or [])
        and not invalid_type
    )


def show_yaxis_ranges(**inputs):
    """
    Boolean function to determine whether "Y-Axis Range" inputs should be displayed or not
    """
    chart_type, y = (inputs.get(p) for p in ["chart_type", "y"])
    return chart_type in YAXIS_CHARTS and len(y or [])


def get_yaxis_type_tabs(y):
    tabs = [
        build_tab("Default", "default", {"padding": "2px", "minWidth": "4em"}),
        build_tab("Single", "single", {"padding": "2px", "minWidth": "4em"}),
    ]
    if len(y) <= 1:
        return tabs
    return tabs + [build_tab("Multi", "multi", {"padding": "2px", "minWidth": "4em"})]


def build_group_val_options(df, group_cols):
    group_vals = find_group_vals(df, group_cols)
    return [
        build_option(
            json.dumps(gv), "|".join([str(gv.get(p, "NaN")) for p in group_cols])
        )
        for gv in group_vals
    ]


def build_map_type_tabs(map_type):
    def _build_hoverable():
        for t in MAP_TYPES:
            if t.get("image", False):
                yield html.Div(
                    [
                        html.Span(t.get("label", t["value"].capitalize())),
                        html.Img(src=build_img_src(t["value"], img_type="map_type")),
                    ],
                    className="col-md-6",
                )

    return html.Div(
        [
            dcc.Tabs(
                id="map-type-tabs",
                value=map_type or "choropleth",
                children=[
                    build_tab(t.get("label", t["value"].capitalize()), t["value"])
                    for t in MAP_TYPES
                ],
                style=dict(height="36px"),
            ),
            html.Div(
                html.Div(list(_build_hoverable()), className="row"),
                className="hoverable__content map-types",
            ),
        ],
        style=dict(paddingLeft=15, borderBottom="none", width="20em"),
        className="hoverable",
    )


def main_inputs_and_group_val_display(inputs):
    if show_group_input(inputs):
        return dict(display="block"), "col-md-8"
    return dict(display="none"), "col-md-12"


def charts_layout(df, settings, **inputs):
    """
    Builds main dash inputs with dropdown options populated with the columns of the dataframe associated with the
    page. Inputs included are: chart tabs, query, x, y, z, group, aggregation, rolling window/computation,
    chart per group toggle, bar sort, bar mode, y-axis range editors

    :param df: dataframe to drive the charts built on page
    :type df: :class:`pandas:pandas.DataFrame`
    :param settings: global settings associated with this dataframe (contains properties like "query")
    :type param: dict
    :return: dash markup
    """
    chart_type, x, y, z, group, agg = (
        inputs.get(p) for p in ["chart_type", "x", "y", "z", "group", "agg"]
    )
    y = y or []
    show_input = show_input_handler(chart_type)
    show_cpg = show_chart_per_group(**inputs)
    show_yaxis = show_yaxis_ranges(**inputs)
    scatter_input = dict(display="block" if chart_type == "scatter" else "none")
    bar_style, barsort_input_style = bar_input_style(**inputs)
    animate_style, animate_by_style, animate_opts = animate_styles(df, **inputs)

    options = build_input_options(df, **inputs)
    (
        x_options,
        y_multi_options,
        y_single_options,
        z_options,
        group_options,
        barsort_options,
        yaxis_options,
    ) = options
    query_placeholder = "Enter pandas query (ex: col1 == 1)"
    query_value = inputs.get("query") or inner_build_query(
        settings, settings.get("query")
    )
    query_label = html.Div(
        [
            html.Span("Query"),
            html.A(
                html.I(className="fa fa-info-circle ml-4"),
                href="https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-query",
                target="_blank",
                style={"color": "white"},
            ),
        ],
        className="input-group-addon",
        style={"minWidth": "7em"},
    )
    yaxis_type = (inputs.get("yaxis") or {}).get("type") or "default"
    yaxis_type_style = (
        {"borderRadius": "0 0.25rem 0.25rem 0"} if yaxis_type == "default" else None
    )
    show_map = chart_type == "maps"
    map_props = ["map_type", "loc_mode", "loc", "lat", "lon", "map_val"]
    map_type, loc_mode, loc, lat, lon, map_val = (inputs.get(p) for p in map_props)
    map_scope, proj, mapbox_style = (
        inputs.get(p) for p in ["scope", "proj", "mapbox_style"]
    )
    loc_options, lat_options, lon_options, map_val_options = build_map_options(
        df, type=map_type, loc=loc, lat=lat, lon=lon, map_val=map_val
    )
    cscale_style = colorscale_input_style(**inputs)
    default_cscale = JET if chart_type == "heatmap" else REDS

    group_val_style, main_input_class = main_inputs_and_group_val_display(inputs)
    group_val = [json.dumps(gv) for gv in inputs.get("group_val") or []]

    def show_style(show):
        return {"display": "block" if show else "none"}

    def show_map_style(show):
        return {} if show else {"display": "none"}

    return html.Div(
        [
            dcc.Store(id="query-data", data=inputs.get("query")),
            dcc.Store(
                id="input-data",
                data={
                    k: v
                    for k, v in inputs.items()
                    if k not in ["cpg", "barmode", "barsort"]
                },
            ),
            dcc.Store(
                id="chart-input-data",
                data={
                    k: v
                    for k, v in inputs.items()
                    if k in ["cpg", "barmode", "barsort"]
                },
            ),
            dcc.Store(
                id="map-input-data",
                data={
                    k: v
                    for k, v in inputs.items()
                    if k
                    in [
                        "map_type",
                        "map_code",
                        "loc_mode",
                        "lat",
                        "lon",
                        "map_val",
                        "scope",
                        "proj",
                    ]
                },
            ),
            dcc.Store(id="range-data"),
            dcc.Store(id="yaxis-data", data=inputs.get("yaxis")),
            dcc.Store(id="last-chart-input-data", data=inputs),
            dcc.Input(id="chart-code", type="hidden"),
            html.Div(
                html.Div(
                    dcc.Tabs(
                        id="chart-tabs",
                        value=chart_type or "line",
                        children=[
                            build_tab(
                                t.get("label", t["value"].capitalize()), t["value"]
                            )
                            for t in CHARTS
                        ],
                        style=dict(height="36px"),
                    ),
                    className="col-md-12",
                ),
                className="row pt-3 pb-3 charts-filters",
            ),
            html.Div(
                html.Div(
                    [
                        html.Div(
                            [
                                query_label,
                                dcc.Input(
                                    id="query-input",
                                    type="text",
                                    placeholder=query_placeholder,
                                    className="form-control",
                                    value=query_value,
                                    style={"lineHeight": "inherit"},
                                ),
                            ],
                            className="input-group mr-3",
                        )
                    ],
                    className="col",
                ),
                className="row pt-3 pb-3 charts-filters",
            ),
            html.Div(
                [
                    html.Div(
                        [
                            html.Div(
                                [
                                    build_input(
                                        [html.Div("X"), html.Small("(Agg By)")],
                                        dcc.Dropdown(
                                            id="x-dropdown",
                                            options=x_options,
                                            placeholder="Select a column",
                                            value=x,
                                            style=dict(width="inherit"),
                                        ),
                                        label_class="input-group-addon d-block pt-1 pb-0",
                                    ),
                                    build_input(
                                        "Y",
                                        dcc.Dropdown(
                                            id="y-multi-dropdown",
                                            options=y_multi_options,
                                            multi=True,
                                            placeholder="Select a column(s)",
                                            style=dict(width="inherit"),
                                            value=y
                                            if show_input("y", "multi")
                                            else None,
                                        ),
                                        className="col",
                                        id="y-multi-input",
                                        style=show_style(show_input("y", "multi")),
                                    ),
                                    build_input(
                                        [html.Div("Y"), html.Small("(Agg By)")],
                                        dcc.Dropdown(
                                            id="y-single-dropdown",
                                            options=y_single_options,
                                            placeholder="Select a column",
                                            style=dict(width="inherit"),
                                            value=y[0]
                                            if show_input("y") and len(y)
                                            else None,
                                        ),
                                        className="col",
                                        label_class="input-group-addon d-block pt-1 pb-0",
                                        id="y-single-input",
                                        style=show_style(show_input("y")),
                                    ),
                                    build_input(
                                        "Z",
                                        dcc.Dropdown(
                                            id="z-dropdown",
                                            options=z_options,
                                            placeholder="Select a column",
                                            style=dict(width="inherit"),
                                            value=z,
                                        ),
                                        className="col",
                                        id="z-input",
                                        style=show_style(show_input("z")),
                                    ),
                                    build_input(
                                        "Group",
                                        dcc.Dropdown(
                                            id="group-dropdown",
                                            options=group_options,
                                            multi=True,
                                            placeholder="Select a group(s)",
                                            value=group,
                                            style=dict(width="inherit"),
                                        ),
                                        className="col",
                                        id="group-input",
                                        style=show_style(show_input("group")),
                                    ),
                                ],
                                id="non-map-inputs",
                                style={} if not show_map else {"display": "none"},
                                className="row p-0 charts-filters",
                            ),
                            html.Div(
                                [
                                    build_map_type_tabs(map_type),
                                    html.Div(
                                        [
                                            html.Div(
                                                [
                                                    build_loc_mode_hover(loc_mode),
                                                    dcc.Dropdown(
                                                        id="map-loc-mode-dropdown",
                                                        options=[
                                                            build_option(
                                                                v,
                                                                LOC_MODE_INFO[v].get(
                                                                    "label"
                                                                ),
                                                            )
                                                            for v in [
                                                                "ISO-3",
                                                                "USA-states",
                                                                "country names",
                                                                "geojson-id",
                                                            ]
                                                        ],
                                                        style=dict(width="inherit"),
                                                        value=loc_mode,
                                                    ),
                                                ],
                                                className="input-group mr-3",
                                            )
                                        ],
                                        id="map-loc-mode-input",
                                        style=show_map_style(map_type == "choropleth"),
                                        className="col-auto",
                                    ),
                                    custom_geojson.build_modal(map_type, loc_mode),
                                    build_input(
                                        [html.Div("Locations"), html.Small("(Agg By)")],
                                        dcc.Dropdown(
                                            id="map-loc-dropdown",
                                            options=loc_options,
                                            placeholder="Select a column",
                                            value=loc,
                                            style=dict(width="inherit"),
                                        ),
                                        id="map-loc-input",
                                        label_class="input-group-addon d-block pt-1 pb-0",
                                        style=show_map_style(map_type == "choropleth"),
                                    ),
                                    build_input(
                                        [html.Div("Lat"), html.Small("(Agg By)")],
                                        dcc.Dropdown(
                                            id="map-lat-dropdown",
                                            options=lat_options,
                                            placeholder="Select a column",
                                            value=lat,
                                            style=dict(width="inherit"),
                                        ),
                                        id="map-lat-input",
                                        label_class="input-group-addon d-block pt-1 pb-0",
                                        style=show_map_style(
                                            map_type in ["scattergeo", "mapbox"]
                                        ),
                                    ),
                                    build_input(
                                        [html.Div("Lon"), html.Small("(Agg By)")],
                                        dcc.Dropdown(
                                            id="map-lon-dropdown",
                                            options=lon_options,
                                            placeholder="Select a column",
                                            style=dict(width="inherit"),
                                            value=lon,
                                        ),
                                        id="map-lon-input",
                                        label_class="input-group-addon d-block pt-1 pb-0",
                                        style=show_map_style(
                                            map_type in ["scattergeo", "mapbox"]
                                        ),
                                    ),
                                    build_input(
                                        "Scope",
                                        dcc.Dropdown(
                                            id="map-scope-dropdown",
                                            options=[build_option(v) for v in SCOPES],
                                            style=dict(width="inherit"),
                                            value=map_scope or "world",
                                        ),
                                        id="map-scope-input",
                                        style=show_map_style(map_type == "scattergeo"),
                                    ),
                                    html.Div(
                                        [
                                            html.Div(
                                                [
                                                    build_mapbox_token_hover(),
                                                    dcc.Dropdown(
                                                        id="map-mapbox-style-dropdown",
                                                        options=build_mapbox_style_options(),
                                                        style=dict(width="inherit"),
                                                        value=mapbox_style
                                                        or "open-street-map",
                                                    ),
                                                ],
                                                className="input-group mr-3",
                                            )
                                        ],
                                        id="map-mapbox-style-input",
                                        className="col-auto",
                                        style=show_map_style(map_type == "mapbox"),
                                    ),
                                    html.Div(
                                        [
                                            html.Div(
                                                [
                                                    build_proj_hover(proj),
                                                    dcc.Dropdown(
                                                        id="map-proj-dropdown",
                                                        options=[
                                                            build_option(v)
                                                            for v in PROJECTIONS
                                                        ],
                                                        style=dict(width="inherit"),
                                                        value=proj,
                                                    ),
                                                ],
                                                className="input-group mr-3",
                                            )
                                        ],
                                        id="map-proj-input",
                                        style=show_map_style(map_type == "scattergeo"),
                                        className="col-auto",
                                    ),
                                    build_input(
                                        "Value",
                                        dcc.Dropdown(
                                            id="map-val-dropdown",
                                            options=map_val_options,
                                            placeholder="Select a column",
                                            style=dict(width="inherit"),
                                            value=map_val,
                                        ),
                                    ),
                                    build_input(
                                        "Group",
                                        dcc.Dropdown(
                                            id="map-group-dropdown",
                                            options=group_options,
                                            multi=True,
                                            placeholder="Select a group(s)",
                                            value=inputs.get("map_group"),
                                            style=dict(width="inherit"),
                                        ),
                                        className="col",
                                        id="map-group-input",
                                    ),
                                ],
                                id="map-inputs",
                                className="row charts-filters",
                                style={} if show_map else {"display": "none"},
                            ),
                            html.Div(
                                [
                                    build_input(
                                        "Aggregation",
                                        dcc.Dropdown(
                                            id="agg-dropdown",
                                            options=[
                                                build_option(v, AGGS[v])
                                                for v in [
                                                    "raw",
                                                    "count",
                                                    "nunique",
                                                    "sum",
                                                    "mean",
                                                    "rolling",
                                                    "corr",
                                                    "first",
                                                    "last",
                                                    "median",
                                                    "min",
                                                    "max",
                                                    "std",
                                                    "var",
                                                    "mad",
                                                    "prod",
                                                    "pctsum",
                                                    "pctct",
                                                ]
                                            ],
                                            placeholder="Select an aggregation",
                                            style=dict(width="inherit"),
                                            value=agg or "raw",
                                        ),
                                    ),
                                    html.Div(
                                        [
                                            build_input(
                                                "Window",
                                                dcc.Input(
                                                    id="window-input",
                                                    type="number",
                                                    placeholder="Enter days",
                                                    className="form-control text-center",
                                                    style={"lineHeight": "inherit"},
                                                    value=inputs.get("window"),
                                                ),
                                            ),
                                            build_input(
                                                "Computation",
                                                dcc.Dropdown(
                                                    id="rolling-comp-dropdown",
                                                    options=[
                                                        build_option(
                                                            "corr", "Correlation"
                                                        ),
                                                        build_option("count", "Count"),
                                                        build_option(
                                                            "cov", "Covariance"
                                                        ),
                                                        build_option(
                                                            "kurt", "Kurtosis"
                                                        ),
                                                        build_option("max", "Maximum"),
                                                        build_option("mean", "Mean"),
                                                        build_option(
                                                            "median", "Median"
                                                        ),
                                                        build_option("min", "Minimum"),
                                                        build_option("skew", "Skew"),
                                                        build_option(
                                                            "std", "Standard Deviation"
                                                        ),
                                                        build_option("sum", "Sum"),
                                                        build_option("var", "Variance"),
                                                    ],
                                                    placeholder="Select an computation",
                                                    style=dict(width="inherit"),
                                                    value=inputs.get("rolling_comp"),
                                                ),
                                            ),
                                        ],
                                        id="rolling-inputs",
                                        style=show_style(agg == "rolling"),
                                    ),
                                ],
                                className="row pt-3 pb-3 charts-filters",
                            ),
                            html.Div(
                                [
                                    build_input(
                                        "Chart Per\nGroup",
                                        html.Div(
                                            daq.BooleanSwitch(
                                                id="cpg-toggle",
                                                on=inputs.get("cpg") or False,
                                            ),
                                            className="toggle-wrapper",
                                        ),
                                        id="cpg-input",
                                        style=show_style(show_cpg),
                                        className="col-auto",
                                    ),
                                    build_input(
                                        "Trendline",
                                        dcc.Dropdown(
                                            id="trendline-dropdown",
                                            options=[
                                                build_option("ols"),
                                                build_option("lowess"),
                                            ],
                                            value=inputs.get("trendline"),
                                        ),
                                        className="col-auto addon-min-width",
                                        style=scatter_input,
                                        id="trendline-input",
                                    ),
                                    build_input(
                                        "Barmode",
                                        dcc.Dropdown(
                                            id="barmode-dropdown",
                                            options=[
                                                build_option("group", "Group"),
                                                build_option("stack", "Stack"),
                                                build_option("relative", "Relative"),
                                            ],
                                            value=inputs.get("barmode") or "group",
                                            placeholder="Select a mode",
                                        ),
                                        className="col-auto addon-min-width",
                                        style=bar_style,
                                        id="barmode-input",
                                    ),
                                    build_input(
                                        "Barsort",
                                        dcc.Dropdown(
                                            id="barsort-dropdown",
                                            options=barsort_options,
                                            value=inputs.get("barsort"),
                                        ),
                                        className="col-auto addon-min-width",
                                        style=barsort_input_style,
                                        id="barsort-input",
                                    ),
                                    html.Div(
                                        html.Div(
                                            [
                                                html.Span(
                                                    "Y-Axis",
                                                    className="input-group-addon",
                                                ),
                                                html.Div(
                                                    dcc.Tabs(
                                                        id="yaxis-type",
                                                        value=yaxis_type,
                                                        children=get_yaxis_type_tabs(y),
                                                    ),
                                                    id="yaxis-type-div",
                                                    className="form-control col-auto pt-3",
                                                    style=yaxis_type_style,
                                                ),
                                                dcc.Dropdown(
                                                    id="yaxis-dropdown",
                                                    options=yaxis_options,
                                                ),
                                                html.Span(
                                                    "Min:",
                                                    className="input-group-addon col-auto",
                                                    id="yaxis-min-label",
                                                ),
                                                dcc.Input(
                                                    id="yaxis-min-input",
                                                    type="number",
                                                    className="form-control col-auto",
                                                    style={"lineHeight": "inherit"},
                                                ),
                                                html.Span(
                                                    "Max:",
                                                    className="input-group-addon col-auto",
                                                    id="yaxis-max-label",
                                                ),
                                                dcc.Input(
                                                    id="yaxis-max-input",
                                                    type="number",
                                                    className="form-control col-auto",
                                                    style={"lineHeight": "inherit"},
                                                ),
                                            ],
                                            className="input-group",
                                            id="yaxis-min-max-options",
                                        ),
                                        className="col-auto addon-min-width",
                                        id="yaxis-input",
                                        style=show_style(show_yaxis),
                                    ),
                                    build_input(
                                        "Colorscale",
                                        dcs.DashColorscales(
                                            id="colorscale-picker",
                                            colorscale=inputs.get("colorscale")
                                            or default_cscale,
                                        ),
                                        className="col-auto addon-min-width",
                                        style=cscale_style,
                                        id="colorscale-input",
                                    ),
                                    build_input(
                                        "Animate",
                                        html.Div(
                                            daq.BooleanSwitch(
                                                id="animate-toggle",
                                                on=inputs.get("animate") or False,
                                            ),
                                            className="toggle-wrapper",
                                        ),
                                        id="animate-input",
                                        style=animate_style,
                                        className="col-auto",
                                    ),
                                    build_input(
                                        "Animate By",
                                        dcc.Dropdown(
                                            id="animate-by-dropdown",
                                            options=animate_opts,
                                            value=inputs.get("animate_by"),
                                        ),
                                        className="col-auto addon-min-width",
                                        style=animate_style,
                                        id="animate-by-input",
                                    ),
                                ],
                                className="row pt-3 pb-5 charts-filters",
                                id="chart-inputs",
                            ),
                        ],
                        id="main-inputs",
                        className=main_input_class,
                    ),
                    build_input(
                        "Group(s)",
                        dcc.Dropdown(
                            id="group-val-dropdown",
                            multi=True,
                            placeholder="Select a group value(s)",
                            value=group_val,
                            style=dict(width="inherit"),
                        ),
                        className="col-md-4 pt-3 pb-5",
                        id="group-val-input",
                        style=group_val_style,
                    ),
                ],
                className="row",
            ),
            dcc.Loading(
                html.Div(id="chart-content", style={"height": "69vh"}), type="circle"
            ),
            dcc.Textarea(id="copy-text", style=dict(position="absolute", left="-110%")),
        ],
        className="charts-body",
    )



import re

import numpy as np
import pandas as pd
from six import string_types

import dtale.global_state as global_state
from dtale.utils import classify_type, find_dtype


class ColumnReplacement(object):
    def __init__(self, data_id, col, replacement_type, cfg, name=None):
        self.data_id = data_id
        if replacement_type == "spaces":
            self.builder = SpaceReplacement(col, cfg, name)
        elif replacement_type == "strings":
            self.builder = StringReplacement(col, cfg, name)
        elif replacement_type == "value":
            self.builder = ValueReplacement(col, cfg, name)
        elif replacement_type == "imputer":  # iterative, knn, simple
            self.builder = ImputerReplacement(col, cfg, name)
        else:
            raise NotImplementedError(
                "'{}' replacement not implemented yet!".format(replacement_type)
            )

    def build_replacements(self):
        return self.builder.build_column(global_state.get_data(self.data_id))

    def build_code(self):
        return self.builder.build_code(global_state.get_data(self.data_id))


def get_inner_replacement_value(val):
    return np.nan if isinstance(val, string_types) and val.lower() == "nan" else val


def get_replacement_value(cfg, prop):
    value = (cfg or {}).get(prop) or "nan"
    return get_inner_replacement_value(value)


def get_inner_replacement_value_as_str(val, series):
    if isinstance(val, string_types) and val.lower() == "nan":
        return "np.nan"
    if classify_type(find_dtype(series)) == "S":
        return "'{value}'".format(value=val)
    return val


def get_replacement_value_as_str(cfg, prop, series):
    value = (cfg or {}).get(prop) or "nan"
    return get_inner_replacement_value_as_str(value, series)


class SpaceReplacement(object):
    def __init__(self, col, cfg, name):
        self.col = col
        self.cfg = cfg
        self.name = name

    def build_column(self, data):
        value = get_replacement_value(self.cfg, "value")
        return data[self.col].replace(r"^\s+$", value, regex=True)

    def build_code(self, data):
        value = get_replacement_value_as_str(self.cfg, "value", data[self.col])
        return "df.loc[:, '{name}'] = df['{col}'].replace(r'^\\s+$', {value}, regex=True)".format(
            name=self.name or self.col, col=self.col, value=value
        )


class StringReplacement(object):
    def __init__(self, col, cfg, name):
        self.col = col
        self.cfg = cfg
        self.name = name

    def parse_cfg(self):
        return (self.cfg[p] for p in ["value", "ignoreCase", "isChar"])

    def build_column(self, data):
        value, ignore_case, is_char = self.parse_cfg()
        flags = re.UNICODE
        if ignore_case:
            flags |= re.IGNORECASE
        value = re.escape(value)
        if is_char:
            value = "[{value}]+".format(value=value)
        regex_pat = re.compile(r"^ *{value} *$".format(value=value), flags=flags)
        replace_with = get_replacement_value(self.cfg, "replace")
        return data[self.col].replace(regex_pat, replace_with, regex=True)

    def build_code(self, data):
        value, ignore_case, is_char = self.parse_cfg()
        flags = re.UNICODE
        if ignore_case:
            flags |= re.IGNORECASE

        regex_exp = "r'^ *{value} *$'.format(value=re.escape({value}))"
        if is_char:
            regex_exp = "r'^ *[{value}]+ *$'.format(value=re.escape({value}))"
        regex_exp = regex_exp.format(value=value)

        replace_with = get_replacement_value_as_str(self.cfg, "replace", data[self.col])

        return (
            "import re\n\n"
            "regex_pat = re.compile({regex_exp}, flags={flags})\n"
            "df.loc[:, '{name}'] = df['{col}'].replace(regex_pat, {replace}, regex=True)"
        ).format(
            name=self.name or self.col,
            col=self.col,
            regex_exp=regex_exp,
            flags=flags,
            replace=replace_with,
        )


class ValueReplacement(object):
    def __init__(self, col, cfg, name):
        self.col = col
        self.cfg = cfg
        self.name = name

    def build_column(self, data):
        s = data[self.col]
        replacements = {}
        col_replacements = []
        for replacement in self.cfg.get("value", []):
            value = get_replacement_value(replacement, "value")
            replacement_type = replacement.get("type")
            if replacement_type == "agg":
                replace = getattr(s, replacement["replace"])()  # min, max, mean, median
                if pd.isnull(replace):
                    raise Exception(
                        "Running the aggregation, {agg}, on {col} resulted in nan, this would result in a no-op."
                    )
            elif replacement_type == "col":
                col_replacements.append(
                    lambda s2: np.where(s2 == value, data[replacement["replace"]], s2)
                )
            else:
                replace = get_replacement_value(replacement, "replace")
            replacements[value] = replace
        final_s = s
        if len(replacements):
            final_s = final_s.replace(replacements)
        for col_r in col_replacements:
            final_s = col_r(final_s)
        return final_s

    def build_code(self, data):
        replacements = []
        series = data[self.col]
        col_replacements = []
        for replacement in self.cfg.get("value", []):
            value = get_replacement_value_as_str(replacement, "value", series)
            replacement_type = self.cfg.get("type")
            if replacement_type == "agg":
                replace = "getattr(df['{col}'], '{agg}')()".format(
                    agg=replacement["value"], col=self.col
                )
            elif replacement_type == "col":
                col_replacements.append(
                    "s = np.where(s == {value}, data['{col2}'], s)".format(
                        col2=replacement["replace"], value=value
                    )
                )
            else:
                replace = get_replacement_value_as_str(replacement, "replace", series)
            replacements.append(
                "\t{value}: {replace}".format(value=value, replace=replace)
            )

        code = ["s = df['{col}']".format(col=self.col)]
        if len(replacements):
            replacements = ",\n".join(replacements)
            replacements = "{\n" + replacements + "}"
            code.append(
                "s = s.replace({replacements})".format(replacements=replacements)
            )
        code += col_replacements
        code.append("df.loc[:, '{name}'] = s".format(name=self.name or self.col))
        return "\n".join(code)


class ImputerReplacement(object):
    def __init__(self, col, cfg, name):
        self.col = col
        self.cfg = cfg
        self.name = name

    def build_column(self, data):
        imputer_type = self.cfg["type"]
        if imputer_type == "iterative":
            try:
                from sklearn.experimental import enable_iterative_imputer  # noqa
                from sklearn.impute import IterativeImputer
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.21.0 installed in order to use the Iterative Imputer!"
                )
            imputer = IterativeImputer()
        elif imputer_type == "knn":
            try:
                from sklearn.impute import KNNImputer
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.22.0 installed in order to use the Iterative Imputer!"
                )
            n_neighbors = self.cfg.get("n_neighbors") or 2
            imputer = KNNImputer(n_neighbors=n_neighbors)
        elif imputer_type == "simple":
            try:
                from sklearn.impute import SimpleImputer
            except ImportError:
                raise Exception(
                    "You must have at least scikit-learn 0.20.0 installed in order to use the Iterative Imputer!"
                )
            imputer = SimpleImputer()
        else:
            raise NotImplementedError(
                "'{}' sklearn imputer not implemented yet!".format(imputer_type)
            )
        output = imputer.fit_transform(data[[self.col]])
        return pd.DataFrame(output, columns=[self.col], index=data.index)[self.col]

    def build_code(self, _data):
        imputer_type = self.cfg["type"]
        code = []
        if imputer_type == "iterative":
            code.append(
                (
                    "from sklearn.experimental import enable_iterative_imputer\n"
                    "from sklearn.impute import IterativeImputer\n\n"
                    "output = IterativeImputer().fit_transform(df[['{col}']])"
                ).format(col=self.col)
            )
        elif imputer_type == "knn":
            n_neighbors = self.cfg.get("n_neighbors") or 2
            code.append(
                (
                    "from sklearn.impute import KNNImputer\n\n"
                    "output = KNNImputer(n_neighbors={n_neighbors}).fit_transform(df[['{col}']])"
                ).format(col=self.col, n_neighbors=n_neighbors)
            )
        elif imputer_type == "simple":
            code.append(
                (
                    "from sklearn.impute import SimpleImputer\n\n"
                    "output = SimpleImputer().fit_transform(df[['{col}']])"
                ).format(col=self.col)
            )
        code.append(
            "df.loc[:, '{name}'] = pd.DataFrame(output, columns=['{col}'], index=df.index)['{col}']".format(
                name=self.name or self.col, col=self.col
            )
        )
        return "\n".join(code)



import json as json
from logging import getLogger

import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

import dtale.dash_application.custom_geojson as custom_geojson
import dtale.global_state as global_state
from dtale.charts.utils import MAX_GROUPS, ZAXIS_CHARTS
from dtale.dash_application.charts import build_chart, chart_url_params
from dtale.dash_application.layout.layout import (
    animate_styles,
    bar_input_style,
    base_layout,
    build_group_val_options,
    build_input_options,
    build_loc_mode_hover_children,
    build_map_options,
    build_mapbox_style_options,
    build_proj_hover_children,
    charts_layout,
    colorscale_input_style,
    get_yaxis_type_tabs,
    main_inputs_and_group_val_display,
    show_chart_per_group,
    show_group_input,
    show_input_handler,
    show_yaxis_ranges,
)
from dtale.utils import dict_merge, is_app_root_defined, make_list, run_query

logger = getLogger(__name__)


class DtaleDash(dash.Dash):
    """
    Wrapper class to dash.Dash to allow for abstraction of global state used for building the default layout.
    Additional state include stylesheets, JS files and styling for github demos.
    """

    def __init__(self, *args, **kwargs):
        server = kwargs.get("server")
        kwargs["external_stylesheets"] = ["/css/main.css", "/css/dash.css"]
        if server.config["GITHUB_FORK"]:
            kwargs["external_stylesheets"].append("/css/github_fork.css")
        kwargs["external_scripts"] = [
            "/dash/components_bundle.js",
            "/dash/custom_bundle.js",
            "/dist/base_styles_bundle.js",
        ]

        app_root = server.config.get("APPLICATION_ROOT")
        if is_app_root_defined(app_root):

            def _prepend_app_root(v):
                return "{}{}".format(app_root, v)

            kwargs["requests_pathname_prefix"] = _prepend_app_root(
                kwargs["routes_pathname_prefix"]
            )
            kwargs["external_stylesheets"] = [
                _prepend_app_root(v) for v in kwargs["external_stylesheets"]
            ]
            kwargs["external_scripts"] = [
                _prepend_app_root(v) for v in kwargs["external_scripts"]
            ]
            kwargs["assets_url_path"] = _prepend_app_root("")
            kwargs["assets_external_path"] = _prepend_app_root("/assets")

        super(DtaleDash, self).__init__(*args, **kwargs)

    def interpolate_index(self, **kwargs):
        return base_layout(
            self.server.config["GITHUB_FORK"],
            self.server.config.get("APPLICATION_ROOT"),
            **kwargs
        )


def add_dash(server):
    """
    Adds dash support to main D-Tale Flask process.

    :param server: main D-Tale Flask process
    :type server: :class:`flask:flask.Flask`
    :return: server with dash added
    :rtype: :class:`flask:flask.Flask`
    """

    dash_app = DtaleDash(
        server=server, routes_pathname_prefix="/charts/", eager_loading=True
    )

    # Since we're adding callbacks to elements that don't exist in the app.layout,
    # Dash will raise an exception to warn us that we might be
    # doing something wrong.
    # In this case, we're adding the elements through a callback, so we can ignore
    # the exception.
    dash_app.config.suppress_callback_exceptions = True
    dash_app.layout = html.Div(
        [dcc.Location(id="url", refresh=False), html.Div(id="popup-content")]
    )
    dash_app.scripts.config.serve_locally = True
    dash_app.css.config.serve_locally = True

    init_callbacks(dash_app)

    return dash_app.server


def get_data_id(pathname):
    """
    Parses data ID from query path (ex: 'foo/bar/1' => '1')
    """
    return global_state.find_data_id(pathname.split("/")[-1])


def init_callbacks(dash_app):
    """
    Dynamically adds dash callbacks to dash-wrapped flask server

    :param dash_app: flask server with dash support enabled
    :type dash_app: :class:`flask:flask.Flask`
    :return: flask server with dash callbacks added
    :rtype: :class:`flask:flask.Flask`
    """

    @dash_app.callback(
        [
            Output("query-data", "data"),
            Output("query-input", "style"),
            Output("query-input", "title"),
        ],
        [Input("query-input", "value")],
        [State("url", "pathname"), State("query-data", "data")],
    )
    def query_input(query, pathname, curr_query):
        """
        dash callback for storing valid pandas dataframe queries.  This acts as an intermediary between values typed
        by the user and values that are applied to pandas dataframes.  Most of the time what the user has typed is not
        complete and thus not a valid pandas dataframe query.

        :param query: query input
        :type query: str
        :param pathname: URL path
        :param curr_query: current valid pandas dataframe query
        :return: tuple of (query (if valid), styling for query input (if invalid input), query input title (containing
        invalid query exception information)
        :rtype: tuple of (str, str, str)
        """
        try:
            data_id = get_data_id(pathname)
            data = global_state.get_data(data_id)
            ctxt_vars = global_state.get_context_variables(data_id)
            run_query(data, query, ctxt_vars)
            return query, {"line-height": "inherit"}, ""
        except BaseException as ex:
            return (
                curr_query,
                {"line-height": "inherit", "background-color": "pink"},
                str(ex),
            )

    @dash_app.callback(
        [
            Output("input-data", "data"),
            Output("x-dropdown", "options"),
            Output("y-single-dropdown", "options"),
            Output("y-multi-dropdown", "options"),
            Output("z-dropdown", "options"),
            Output("group-dropdown", "options"),
            Output("barsort-dropdown", "options"),
            Output("yaxis-dropdown", "options"),
            Output("non-map-inputs", "style"),
            Output("map-inputs", "style"),
            Output("colorscale-input", "style"),
        ],
        [
            Input("query-data", "modified_timestamp"),
            Input("chart-tabs", "value"),
            Input("x-dropdown", "value"),
            Input("y-multi-dropdown", "value"),
            Input("y-single-dropdown", "value"),
            Input("z-dropdown", "value"),
            Input("group-dropdown", "value"),
            Input("group-val-dropdown", "value"),
            Input("agg-dropdown", "value"),
            Input("window-input", "value"),
            Input("rolling-comp-dropdown", "value"),
        ],
        [State("url", "pathname"), State("query-data", "data")],
    )
    def input_data(
        _ts,
        chart_type,
        x,
        y_multi,
        y_single,
        z,
        group,
        group_val,
        agg,
        window,
        rolling_comp,
        pathname,
        query,
    ):
        """
        dash callback for maintaining chart input state and column-based dropdown options.  This will guard against
        users selecting the same column for multiple axes.
        """
        y_val = make_list(y_single if chart_type in ZAXIS_CHARTS else y_multi)
        if group_val is not None:
            group_val = [json.loads(gv) for gv in group_val]
        inputs = dict(
            query=query,
            chart_type=chart_type,
            x=x,
            y=y_val,
            z=z,
            group=group,
            group_val=group_val,
            agg=agg,
            window=window,
            rolling_comp=rolling_comp,
        )
        data_id = get_data_id(pathname)
        options = build_input_options(global_state.get_data(data_id), **inputs)
        (
            x_options,
            y_multi_options,
            y_single_options,
            z_options,
            group_options,
            barsort_options,
            yaxis_options,
        ) = options
        show_map = chart_type == "maps"
        map_style = {} if show_map else {"display": "none"}
        non_map_style = {"display": "none"} if show_map else {}
        cscale_style = colorscale_input_style(chart_type=chart_type)
        return (
            inputs,
            x_options,
            y_single_options,
            y_multi_options,
            z_options,
            group_options,
            barsort_options,
            yaxis_options,
            non_map_style,
            map_style,
            cscale_style,
        )

    @dash_app.callback(
        [
            Output("map-input-data", "data"),
            Output("map-loc-dropdown", "options"),
            Output("map-lat-dropdown", "options"),
            Output("map-lon-dropdown", "options"),
            Output("map-val-dropdown", "options"),
            Output("map-loc-mode-input", "style"),
            Output("map-loc-input", "style"),
            Output("map-lat-input", "style"),
            Output("map-lon-input", "style"),
            Output("map-scope-input", "style"),
            Output("map-mapbox-style-input", "style"),
            Output("map-proj-input", "style"),
            Output("proj-hover", "style"),
            Output("proj-hover", "children"),
            Output("loc-mode-hover", "style"),
            Output("loc-mode-hover", "children"),
            Output("custom-geojson-input", "style"),
        ],
        [
            Input("map-type-tabs", "value"),
            Input("map-loc-mode-dropdown", "value"),
            Input("map-loc-dropdown", "value"),
            Input("map-lat-dropdown", "value"),
            Input("map-lon-dropdown", "value"),
            Input("map-val-dropdown", "value"),
            Input("map-scope-dropdown", "value"),
            Input("map-mapbox-style-dropdown", "value"),
            Input("map-proj-dropdown", "value"),
            Input("map-group-dropdown", "value"),
            Input("geojson-dropdown", "value"),
            Input("featureidkey-dropdown", "value"),
        ],
        [State("url", "pathname")],
    )
    def map_data(
        map_type,
        loc_mode,
        loc,
        lat,
        lon,
        map_val,
        scope,
        style,
        proj,
        group,
        geojson,
        featureidkey,
        pathname,
    ):
        data_id = get_data_id(pathname)
        map_type = map_type or "choropleth"
        if map_type == "choropleth":
            map_data = dict(
                map_type=map_type, loc_mode=loc_mode, loc=loc, map_val=map_val
            )
            if loc_mode == "geojson-id":
                map_data["geojson"] = geojson
                map_data["featureidkey"] = featureidkey
        elif map_type == "mapbox":
            map_data = dict(
                map_type=map_type, lat=lat, lon=lon, map_val=map_val, mapbox_style=style
            )
        else:
            map_data = dict(
                map_type=map_type,
                lat=lat,
                lon=lon,
                map_val=map_val,
                scope=scope,
                proj=proj,
            )

        if group is not None:
            map_data["map_group"] = group
        df = global_state.get_data(data_id)
        loc_options, lat_options, lon_options, map_val_options = build_map_options(
            df, type=map_type, loc=loc, lat=lat, lon=lon, map_val=map_val
        )
        choro_style = {} if map_type == "choropleth" else {"display": "none"}
        coord_style = (
            {} if map_type in ["scattergeo", "mapbox"] else {"display": "none"}
        )
        scatt_style = {} if map_type == "scattergeo" else {"display": "none"}
        mapbox_style = {} if map_type == "mapbox" else {"display": "none"}
        proj_hover_style = (
            {"display": "none"} if proj is None else dict(borderBottom="none")
        )
        proj_hopver_children = build_proj_hover_children(proj)
        loc_mode_hover_style = (
            {"display": "none"} if loc_mode is None else dict(borderBottom="none")
        )
        loc_mode_children = build_loc_mode_hover_children(loc_mode)
        custom_geojson_link = (
            {}
            if map_type == "choropleth" and loc_mode == "geojson-id"
            else {"display": "none"}
        )
        return (
            map_data,
            loc_options,
            lat_options,
            lon_options,
            map_val_options,
            choro_style,
            choro_style,
            coord_style,
            coord_style,
            scatt_style,
            mapbox_style,
            scatt_style,
            proj_hover_style,
            proj_hopver_children,
            loc_mode_hover_style,
            loc_mode_children,
            custom_geojson_link,
        )

    @dash_app.callback(
        Output("map-mapbox-style-dropdown", "options"),
        [Input("mapbox-token-input", "value")],
    )
    def update_mapbox_token(token):
        from dtale.charts.utils import set_mapbox_token

        if token:
            set_mapbox_token(token)
        return build_mapbox_style_options()

    @dash_app.callback(
        [
            Output("y-multi-input", "style"),
            Output("y-single-input", "style"),
            Output("z-input", "style"),
            Output("group-input", "style"),
            Output("rolling-inputs", "style"),
            Output("cpg-input", "style"),
            Output("barmode-input", "style"),
            Output("barsort-input", "style"),
            Output("yaxis-input", "style"),
            Output("animate-input", "style"),
            Output("animate-by-input", "style"),
            Output("animate-by-dropdown", "options"),
            Output("trendline-input", "style"),
        ],
        [Input("input-data", "modified_timestamp")],
        [State("input-data", "data"), State("url", "pathname")],
    )
    def input_toggles(_ts, inputs, pathname):
        """
        dash callback controlling showing/hiding of chart-specific inputs (for example z-axis) as well as chart
        formatting inputs (sorting for bars in bar chart, bar chart style (stacked) or y-axis ranges.
        """
        [chart_type, agg] = [inputs.get(p) for p in ["chart_type", "agg"]]
        show_input = show_input_handler(chart_type)

        y_multi_style = {"display": "block" if show_input("y", "multi") else "none"}
        y_single_style = {"display": "block" if show_input("y") else "none"}
        z_style = {"display": "block" if show_input("z") else "none"}
        group_style = {"display": "block" if show_input("group") else "none"}
        rolling_style = {"display": "inherit" if agg == "rolling" else "none"}
        cpg_style = {"display": "block" if show_chart_per_group(**inputs) else "none"}
        bar_style, barsort_style = bar_input_style(**inputs)
        yaxis_style = {"display": "block" if show_yaxis_ranges(**inputs) else "none"}

        data_id = get_data_id(pathname)
        df = global_state.get_data(data_id)
        animate_style, animate_by_style, animate_opts = animate_styles(df, **inputs)
        trendline_style = dict(display="block" if chart_type == "scatter" else "none")
        return (
            y_multi_style,
            y_single_style,
            z_style,
            group_style,
            rolling_style,
            cpg_style,
            bar_style,
            barsort_style,
            yaxis_style,
            animate_style,
            animate_by_style,
            animate_opts,
            trendline_style,
        )

    @dash_app.callback(
        Output("chart-input-data", "data"),
        [
            Input("cpg-toggle", "on"),
            Input("barmode-dropdown", "value"),
            Input("barsort-dropdown", "value"),
            Input("colorscale-picker", "colorscale"),
            Input("animate-toggle", "on"),
            Input("animate-by-dropdown", "value"),
            Input("trendline-dropdown", "value"),
        ],
    )
    def chart_input_data(
        cpg, barmode, barsort, colorscale, animate, animate_by, trendline
    ):
        """
        dash callback for maintaining selections in chart-formatting inputs
            - chart per group flag
            - bar chart mode
            - bar chart sorting
        """
        return dict(
            cpg=cpg,
            barmode=barmode,
            barsort=barsort,
            colorscale=colorscale,
            animate=animate,
            animate_by=animate_by,
            trendline=trendline,
        )

    @dash_app.callback(
        [
            Output("chart-content", "children"),
            Output("last-chart-input-data", "data"),
            Output("range-data", "data"),
            Output("chart-code", "value"),
            Output("yaxis-type", "children"),
        ],
        # Since we use the data prop in an output,
        # we cannot get the initial data on load with the data prop.
        # To counter this, you can use the modified_timestamp
        # as Input and the data as State.
        # This limitation is due to the initial None callbacks
        # https://github.com/plotly/dash-renderer/pull/81
        [
            Input("input-data", "modified_timestamp"),
            Input("chart-input-data", "modified_timestamp"),
            Input("yaxis-data", "modified_timestamp"),
            Input("map-input-data", "modified_timestamp"),
        ],
        [
            State("url", "pathname"),
            State("input-data", "data"),
            State("chart-input-data", "data"),
            State("yaxis-data", "data"),
            State("map-input-data", "data"),
            State("last-chart-input-data", "data"),
        ],
    )
    def on_data(
        _ts1,
        _ts2,
        _ts3,
        _ts4,
        pathname,
        inputs,
        chart_inputs,
        yaxis_data,
        map_data,
        last_chart_inputs,
    ):
        """
        dash callback controlling the building of dash charts
        """
        all_inputs = dict_merge(
            inputs, chart_inputs, dict(yaxis=yaxis_data or {}), map_data
        )
        if all_inputs == last_chart_inputs:
            raise PreventUpdate
        if is_app_root_defined(dash_app.server.config.get("APPLICATION_ROOT")):
            all_inputs["app_root"] = dash_app.server.config["APPLICATION_ROOT"]
        charts, range_data, code = build_chart(get_data_id(pathname), **all_inputs)
        return (
            charts,
            all_inputs,
            range_data,
            code,
            get_yaxis_type_tabs(make_list(inputs.get("y") or [])),
        )

    def get_default_range(range_data, y, max=False):
        if max:
            return next(
                iter(
                    sorted(
                        [range_data[y2] for y2 in y if y2 in range_data], reverse=True
                    )
                ),
                None,
            )
        return next(
            iter(sorted([range_data[y2] for y2 in y if y2 in range_data])), None
        )

    @dash_app.callback(
        [
            Output("yaxis-min-input", "value"),
            Output("yaxis-max-input", "value"),
            Output("yaxis-dropdown", "style"),
            Output("yaxis-min-label", "style"),
            Output("yaxis-min-input", "style"),
            Output("yaxis-max-label", "style"),
            Output("yaxis-max-input", "style"),
            Output("yaxis-type-div", "style"),
        ],
        [Input("yaxis-type", "value"), Input("yaxis-dropdown", "value")],
        [
            State("input-data", "data"),
            State("yaxis-data", "data"),
            State("range-data", "data"),
        ],
    )
    def yaxis_min_max_values(yaxis_type, yaxis, inputs, yaxis_inputs, range_data):
        """
        dash callback controlling values for selected y-axis in y-axis range editor
        """
        y = make_list(inputs.get("y"))
        dd_style = dict(
            display="block" if yaxis_type == "multi" and len(y) > 1 else "none"
        )
        type_style = (
            {"borderRadius": "0 0.25rem 0.25rem 0"} if yaxis_type == "default" else None
        )
        min_max_style = (
            "none"
            if (yaxis_type == "default") or (yaxis_type == "multi" and yaxis is None)
            else "block"
        )
        label_style = dict(display=min_max_style)
        input_style = {"lineHeight": "inherit", "display": min_max_style}
        curr_min, curr_max = (None, None)
        range_min, range_max = ((range_data or {}).get(p) or {} for p in ["min", "max"])
        if yaxis:
            curr_vals = (yaxis_inputs or {}).get("data", {}).get(yaxis) or {}
            curr_min = curr_vals.get("min") or range_min.get(yaxis)
            curr_max = curr_vals.get("max") or range_max.get(yaxis)
        elif yaxis_type == "single":
            curr_vals = (yaxis_inputs or {}).get("data", {}).get("all") or {}
            curr_min = curr_vals.get("min")
            if curr_min is None:
                curr_min = get_default_range(range_min, y)
            curr_max = curr_vals.get("max")
            if curr_max is None:
                curr_max = get_default_range(range_max, y, max=True)
        return (
            curr_min,
            curr_max,
            dd_style,
            label_style,
            input_style,
            label_style,
            input_style,
            type_style,
        )

    @dash_app.callback(
        Output("yaxis-data", "data"),
        [
            Input("yaxis-type", "value"),
            Input("yaxis-min-input", "value"),
            Input("yaxis-max-input", "value"),
        ],
        [
            State("yaxis-dropdown", "value"),
            State("yaxis-data", "data"),
            State("range-data", "data"),
            State("input-data", "data"),
        ],
    )
    def update_yaxis_data(
        yaxis_type, yaxis_min, yaxis_max, yaxis, yaxis_data, range_data, inputs
    ):
        """
        dash callback controlling updates to y-axis range state
        """
        yaxis_data = yaxis_data or dict(data={})
        yaxis_data["type"] = yaxis_type
        yaxis_name = "all" if yaxis_type == "single" else yaxis
        if yaxis_name == "all":
            y = make_list(inputs.get("y"))
            mins = range_data.get("min", {})
            maxs = range_data.get("max", {})
            range_min = get_default_range(mins, y)
            range_max = get_default_range(maxs, y, max=True)
        elif yaxis is None:
            raise PreventUpdate
        else:
            range_min, range_max = (
                range_data[p].get(yaxis_name) for p in ["min", "max"]
            )

        if yaxis_name in yaxis_data["data"]:
            if (yaxis_min, yaxis_max) == (range_min, range_max):
                del yaxis_data["data"][yaxis_name]
            else:
                yaxis_data["data"][yaxis_name] = dict(min=yaxis_min, max=yaxis_max)
        else:
            if (yaxis_min, yaxis_max) != (range_min, range_max):
                yaxis_data["data"][yaxis_name] = dict(min=yaxis_min, max=yaxis_max)
        return yaxis_data

    @dash_app.callback(
        [Output("group-val-input", "style"), Output("main-inputs", "className")],
        [
            Input("input-data", "modified_timestamp"),
            Input("map-input-data", "modified_timestamp"),
        ],
        [State("input-data", "data"), State("map-input-data", "data")],
    )
    def main_input_class(ts_, ts2_, inputs, map_inputs):
        return main_inputs_and_group_val_display(dict_merge(inputs, map_inputs))

    @dash_app.callback(
        [
            Output("group-val-dropdown", "options"),
            Output("group-val-dropdown", "value"),
        ],
        [
            Input("chart-tabs", "value"),
            Input("group-dropdown", "value"),
            Input("map-group-dropdown", "value"),
        ],
        [
            State("url", "pathname"),
            State("input-data", "data"),
            State("group-val-dropdown", "value"),
        ],
    )
    def group_values(
        chart_type, group_cols, map_group_cols, pathname, inputs, prev_group_vals
    ):
        group_cols = make_list(map_group_cols if chart_type == "maps" else group_cols)
        if not show_group_input(inputs, group_cols):
            return [], None
        data_id = get_data_id(pathname)
        group_vals = run_query(
            global_state.get_data(data_id),
            inputs.get("query"),
            global_state.get_context_variables(data_id),
        )
        group_vals = build_group_val_options(group_vals, group_cols)
        selections = []
        available_vals = [gv["value"] for gv in group_vals]
        if prev_group_vals is not None:
            selections = [pgv for pgv in prev_group_vals if pgv in available_vals]
        if not len(selections) and len(group_vals) <= MAX_GROUPS:
            selections = available_vals

        return group_vals, selections

    @dash_app.callback(
        Output("popup-content", "children"),
        [Input("url", "pathname"), Input("url", "search")],
    )
    def display_page(pathname, search):
        """
        dash callback which gets called on initial load of each dash page (main & popup)
        """
        dash_app.config.suppress_callback_exceptions = False
        if pathname is None:
            raise PreventUpdate
        params = chart_url_params(search)
        data_id = get_data_id(pathname)
        df = global_state.get_data(data_id)
        settings = global_state.get_settings(data_id) or {}
        return charts_layout(df, settings, **params)

    custom_geojson.init_callbacks(dash_app)



import os
import platform
from logging import getLogger

import click

from dtale.cli.loaders import arctic_loader, csv_loader, json_loader

logger = getLogger(__name__)


def build_custom_module_loader_args(fname, path):
    return "dtale.cli.loaders.{}".format(fname), "{}/{}.py".format(path, fname)


def get_py35_loader(fname, path):
    """
    Utility function for loading dynamic modules (CLI configurations) when python_version >= 3.5

    """
    import importlib.util

    spec = importlib.util.spec_from_file_location(
        *build_custom_module_loader_args(fname, path)
    )
    custom_loader = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(custom_loader)
    return custom_loader


def get_py33_loader(fname, path):
    """
    Utility function for loading dynamic modules (CLI configurations) when python_version in (3.3, 3.4)

    """
    from importlib.machinery import SourceFileLoader

    return SourceFileLoader(*build_custom_module_loader_args(fname, path)).load_module()


def get_py2_loader(fname, path):
    """
    Utility function for loading dynamic modules (CLI configurations) when python_version < 3

    """
    import imp

    return imp.load_source(*build_custom_module_loader_args(fname, path))


def unsupported_python_version(version_tuple):
    return (
        "Unsupported version of python used for custom CLI loaders, {}. If you do not plan on using any custom "
        "CLI loaders please remove your DTALE_CLI_LOADERS environment variable."
    ).format(version_tuple)


def custom_module_loader():
    """
    Utility function for using different module loaders based on python version:
    * :meth:`dtale.cli.loaders.get_py35_loader`
    * :meth:`dtale.cli.loaders.get_py33_loader`
    * :meth:`dtale.cli.loaders.get_py2_loader`

    """
    major, minor, revision = [int(i) for i in platform.python_version_tuple()]
    if major == 2:
        return get_py2_loader
    if major == 3:
        if minor >= 5:
            return get_py35_loader
        elif minor in (3, 4):
            return get_py33_loader
    raise ValueError(unsupported_python_version(platform.python_version_tuple()))


LOADERS = {
    arctic_loader.LOADER_KEY: arctic_loader,
    csv_loader.LOADER_KEY: csv_loader,
    json_loader.LOADER_KEY: json_loader,
}


def build_loaders():
    """
    Utility function executed at runtime to load dynamic CLI options from the environment variable, DTALE_CLI_LOADERS.
    You either override one of the two default loader configurations, arctic or csv, or create a brand new
    configuration which can be referenced from the command line.

    """
    global LOADERS

    custom_loader_path = os.environ.get("DTALE_CLI_LOADERS")
    if custom_loader_path is not None:
        custom_loader_func = custom_module_loader()
        for full_filename in os.listdir(custom_loader_path):
            filename, file_extension = os.path.splitext(full_filename)
            if file_extension == ".py":
                custom_loader = custom_loader_func(filename, custom_loader_path)
                if hasattr(custom_loader, "LOADER_KEY") and hasattr(
                    custom_loader, "LOADER_PROPS"
                ):
                    LOADERS[custom_loader.LOADER_KEY] = custom_loader


def setup_loader_options():
    """
    Utility function executed at runtime to find dynamic CLI options as well as the defaults and create their
    `click` decorators to keyword arguments will be processed accordingly.

    """
    build_loaders()

    def decorator(f):
        for cli_loader in LOADERS.values():
            if len(cli_loader.LOADER_PROPS):
                for p in cli_loader.LOADER_PROPS:
                    prop_name = "--{}".format(cli_loader.LOADER_KEY)
                    if isinstance(p, str):
                        if len(p):
                            prop_name = "{}-{}".format(prop_name, p)
                        f = click.option(
                            prop_name, help="Override {}".format(prop_name)
                        )(f)
                    elif isinstance(p, dict):
                        p_name = p.get("name")
                        if len(p_name):
                            prop_name = "{}-{}".format(prop_name, p_name)
                        opt_kwargs = {k: v for k, v in p.items() if k != "name"}
                        f = click.option(prop_name, **opt_kwargs)(f)
                    else:
                        raise NotImplementedError(
                            "You have specified an unknown type for a value in {} LOADER_PROPS: {}".format(
                                cli_loader.LOADER_KEY, type(p)
                            )
                        )
            else:
                f = click.option(
                    "--{}".format(cli_loader.LOADER_KEY),
                    is_flag=True,
                    help="Use {} loader".format(cli_loader.LOADER_KEY),
                )(f)
        return f

    return decorator


def check_loaders(kwargs):
    """
    Utility function to find which CLI loader is being used based on the `click` options/flags provided from the
    command line

    """
    for cli_loader in LOADERS.values():
        cli_loader_func = cli_loader.find_loader(kwargs)
        if cli_loader_func is not None:
            return cli_loader_func
    return None



import dash_html_components as html


def build_input(
    label, input, className="col-auto", label_class="input-group-addon", **kwargs
):
    """
    Helper function to build a standard label/input component in dash.

    :param label: name of the input you are displaying
    :type label: str
    :param input: dash component for storing state
    :param className: style class to be applied to encapsulating div
    :type className: str
    :param kwargs: Optional keyword arguments to be applied to encapsulating div (style, title, id...)
    :type kwargs: dict
    :return: dash components for label/input
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    return html.Div(
        [
            html.Div(
                [html.Span(label, className=label_class), input],
                className="input-group mr-3",
            )
        ],
        className=className,
        **kwargs
    )


def build_option(value, label=None):
    """
    Returns value/label inputs in a dictionary for use in
    :dash:`dash_core_components.Dropdown <dash-core-components/Dropdown>`
    """
    return {"label": label or value, "value": value}



import json

import dtale.global_state as global_state
from dtale.utils import classify_type, find_dtype, make_list


class ColumnFilter(object):
    def __init__(self, data_id, column, cfg):
        self.data_id = data_id
        self.column = column
        s = global_state.get_data(data_id)[column]
        dtype = find_dtype(s)
        self.classification = classify_type(dtype)
        self.cfg = cfg
        if self.cfg is not None:
            self.cfg = json.loads(self.cfg)
        if self.cfg["type"] == "string":
            self.builder = StringFilter(column, self.classification, self.cfg)
        if self.cfg["type"] in ["int", "float"]:
            self.builder = NumericFilter(column, self.classification, self.cfg)
        if self.cfg["type"] == "date":
            self.builder = DateFilter(column, self.classification, self.cfg)
        if self.cfg["type"] == "outliers":
            self.builder = OutlierFilter(column, self.classification, self.cfg)

    def save_filter(self):
        curr_settings = global_state.get_settings(self.data_id)
        filters_key = "{}Filters".format(
            "outlier" if self.cfg["type"] == "outliers" else "column"
        )
        curr_filters = curr_settings.get(filters_key) or {}
        fltr = self.builder.build_filter()
        if fltr is None:
            curr_filters.pop(self.column, None)
        else:
            curr_filters[self.column] = fltr
        curr_settings[filters_key] = curr_filters
        global_state.set_settings(self.data_id, curr_settings)
        return curr_filters


class OutlierFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def build_filter(self):
        if self.cfg.get("query") is None:
            return None
        return self.cfg


class MissingFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def handle_missing(self, fltr):
        if self.cfg is None or not self.cfg.get("missing", False):
            return fltr
        return {"missing": True, "query": "{col} != {col}".format(col=self.column)}


class StringFilter(MissingFilter):
    def __init__(self, column, classification, cfg):
        super(StringFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None or not len(self.cfg.get("value", [])):
            return super(StringFilter, self).handle_missing(None)

        state = self.cfg.get("value", [])
        operand = self.cfg.get("operand", "=")
        fltr = dict(value=state)
        if len(state) == 1:
            val_str = ("'{}'" if self.classification == "S" else "{}").format(state[0])
            fltr["query"] = "{} {} {}".format(
                self.column, "==" if operand == "=" else "!=", val_str
            )
        else:
            val_str = (
                "'{}'".format("', '".join(state))
                if self.classification == "S"
                else ",".join(state)
            )
            fltr["query"] = "{} {} ({})".format(
                self.column, "in" if operand == "=" else "not in", val_str
            )
        return super(StringFilter, self).handle_missing(fltr)


class NumericFilter(MissingFilter):
    def __init__(self, column, classification, cfg):
        super(NumericFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(NumericFilter, self).handle_missing(None)
        cfg_val, cfg_operand, cfg_min, cfg_max = (
            self.cfg.get(p) for p in ["value", "operand", "min", "max"]
        )

        if cfg_operand in ["=", "ne"]:
            state = make_list(cfg_val or [])
            if not len(state):
                return super(NumericFilter, self).handle_missing(None)
            fltr = dict(value=cfg_val, operand=cfg_operand)
            if len(state) == 1:
                fltr["query"] = "{} {} {}".format(
                    self.column, "==" if cfg_operand == "=" else "!=", state[0]
                )
            else:
                fltr["query"] = "{} {} ({})".format(
                    self.column,
                    "in" if cfg_operand == "=" else "not in",
                    ", ".join(state),
                )
            return super(NumericFilter, self).handle_missing(fltr)
        if cfg_operand in ["<", ">", "<=", ">="]:
            if cfg_val is None:
                return super(NumericFilter, self).handle_missing(None)
            fltr = dict(
                value=cfg_val,
                operand=cfg_operand,
                query="{} {} {}".format(self.column, cfg_operand, cfg_val),
            )
            return super(NumericFilter, self).handle_missing(fltr)
        if cfg_operand in ["[]", "()"]:
            fltr = dict(operand=cfg_operand)
            queries = []
            if cfg_min is not None:
                fltr["min"] = cfg_min
                queries.append(
                    "{} >{} {}".format(
                        self.column, "=" if cfg_operand == "[]" else "", cfg_min
                    )
                )
            if cfg_max is not None:
                fltr["max"] = cfg_max
                queries.append(
                    "{} <{} {}".format(
                        self.column, "=" if cfg_operand == "[]" else "", cfg_max
                    )
                )
            if len(queries) == 2 and cfg_max == cfg_min:
                queries = ["{} == {}".format(self.column, cfg_max)]
            if not len(queries):
                return super(NumericFilter, self).handle_missing(None)
            fltr["query"] = " and ".join(queries)
            return super(NumericFilter, self).handle_missing(fltr)
        return super(NumericFilter, self).handle_missing(None)


class DateFilter(MissingFilter):
    def __init__(self, column, classification, cfg):
        super(DateFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(DateFilter, self).handle_missing(None)

        start, end = (self.cfg.get(p) for p in ["start", "end"])
        fltr = dict(start=start, end=end)
        queries = []
        if start:
            queries.append("{} >= '{}'".format(self.column, start))
        if end:
            queries.append("{} <= '{}'".format(self.column, end))
        if len(queries) == 2 and start == end:
            queries = ["{} == '{}'".format(self.column, start)]
        if not len(queries):
            return super(DateFilter, self).handle_missing(None)
        fltr["query"] = " and ".join(queries)
        return super(DateFilter, self).handle_missing(fltr)



from flask import Blueprint

dtale = Blueprint("dtale", __name__, url_prefix="/dtale")

# flake8: NOQA
from dtale.app import show, get_instance, instances, offline_chart  # isort:skip
from dtale.cli.loaders import LOADERS  # isort:skip

for loader_name, loader in LOADERS.items():
    if hasattr(loader, "show_loader"):
        globals()["show_{}".format(loader_name)] = loader.show_loader



#!/usr/bin/env python
from __future__ import print_function

import logging
import sys

import six
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

# Convert Markdown to RST for PyPI
# http://stackoverflow.com/a/26737672
try:
    import pypandoc

    long_description = pypandoc.convert("README.md", "rst")
    changelog = pypandoc.convert("CHANGES.md", "rst")
except (IOError, ImportError, OSError):
    long_description = open("README.md").read()
    changelog = open("CHANGES.md").read()


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(name)s %(message)s", level="DEBUG"
        )

        # import here, cause outside the eggs aren't loaded
        import pytest

        args = (
            [self.pytest_args]
            if isinstance(self.pytest_args, six.string_types)
            else list(self.pytest_args)
        )
        args.extend(
            [
                "--cov",
                "dtale",
                "--cov-report",
                "xml",
                "--cov-report",
                "html",
                "--junitxml",
                "junit.xml",
                "-v",
            ]
        )
        errno = pytest.main(args)
        sys.exit(errno)


setup(
    name="dtale",
    version="1.8.18",
    author="MAN Alpha Technology",
    author_email="ManAlphaTech@man.com",
    description="Web Client for Visualizing Pandas Objects",
    license="LGPL",
    long_description="\n".join((long_description, changelog)),
    keywords=["numeric", "pandas", "visualization", "flask"],
    url="https://github.com/man-group/dtale",
    install_requires=[
        "lz4<=2.2.1; python_version < '3.0'",
        "lz4; python_version > '3.0'",
        "dash>=1.5.0",
        "dash-bootstrap-components",
        "dash-colorscales",
        "dash_daq",
        "Flask>=1.0",
        "Flask-Compress",
        "flask-ngrok; python_version > '3.0'",
        "future >= 0.14.0",
        "itsdangerous",
        "pandas",
        "requests",
        "scikit-learn >= '0.21.0",
        "scipy",
        "statsmodels",
        "six",
        "xarray",
    ],
    extras_require={
        "arctic": ["arctic"],
        "r": ["rpy2<=2.8.6; python_version < '3.0'", "rpy2; python_version > '3.0'"],
        "redis": ["redislite"],
        "tests": ["ipython", "mock", "pytest", "pytest-cov", "pytest-server-fixtures"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    cmdclass={"test": PyTest},
    packages=find_packages(exclude=["tests*", "script*"]),
    package_data={
        "dtale": [
            "dash_application/components/*",
            "static/dist/*",
            "static/dash/*",
            "static/css/*",
            "static/fonts/*",
            "static/images/*",
            "static/images/**/*",
            "static/maps/*",
            "templates/**/*",
            "templates/**/**/*",
        ]
    },
    entry_points={"console_scripts": ["dtale = dtale.cli.script:main"]},
    zip_safe=False,
)
